import math
import numpy as np
from tqdm import tqdm
import os, sys
import itertools
import time
import torchvision
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
from torchvision import transforms
from collections import defaultdict

sys.path.append('../')


from src.optimization import SFW, PositiveKSparsePolytope
from barbar import Bar
from torch.distributions import Categorical

from src.basic_nn_classes import Classifier, SimpleNet

from src.models import get_model
from src.utilities import AddMaskChannel, Averager, categorical_accuracy

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

import copy

from src.arthur_merlin_networks import opt_mask_NN_FW, train_mask_epoch, evaluate_mask_model
from src.arthur_merlin_optimizers import opt_mask
from src.arthur_merlin_utilities import set_labels, Metrics_collector, set_labels
from src.basic_training import regular_training, evaluate_model
from src.arthur_merlin_plotting import show_MnM_images, show_example_images

import importlib

import matplotlib
from math import log2

from params_file import *

import argparse
parser = argparse.ArgumentParser()

parser.add_argument('--epocs', help='Number of epochs to run', default=default_epocs, type=int)
parser.add_argument('--bs', help='Batch Size', default=default_bs, type=int)

parser.add_argument('--lr', help='Learning Rate', default=default_lr, type=float)
parser.add_argument('--en', help='Expected Noise', default=default_en, type=float)
parser.add_argument('--mt', help='Model Type', default=default_mt, type=str)

parser.add_argument('--gamma', help='Parameter to tune the objective function', default=default_gamma, type=float)

parser.add_argument('--merlin_nn', help='is merlin a NN', type=int)

args = parser.parse_args()

main_direc = f"./"


torch.cuda.is_available()

batch_size = args.bs
learning_rate = args.lr
expected_noise = args.en
model_type = args.mt
nepochs = args.epocs
gamma = args.gamma


device = torch.device("cpu")

if args.merlin_nn == 1:
    use_NN_merlin = True
else:
    use_NN_merlin = False

model_names = f""
if use_NN_merlin == 1:
    model_names += "NN"
else:
    model_names += "Opt"

use_NN_morgana = None

nclasses = len(classes)+1

params={}
params['expected_noise'] = expected_noise
params['nclasses'] = nclasses
params['device'] = device
params['classes'] = classes

transformChain = transforms.Compose([transforms.ToTensor(), AddMaskChannel()])
trainData = torchvision.datasets.MNIST(root='.', train=True, download=True, transform=transformChain)
testData = torchvision.datasets.MNIST(root='.', train=False, download=True, transform=transformChain)

if restrict_size == True:
    trainData.data = trainData.data[:1000]
    trainData.targets = trainData.targets[:1000]
    
trainData, testData = set_labels(trainData, testData, classes)


trainLoader = torch.utils.data.DataLoader(trainData, batch_size=batch_size, shuffle=True, pin_memory=torch.cuda.is_available(), num_workers=2)
testLoader = torch.utils.data.DataLoader(testData, batch_size=batch_size, shuffle=False, pin_memory=torch.cuda.is_available(), num_workers=2)

loss_criterion = torch.nn.CrossEntropyLoss().to(device=device)
softmax = torch.nn.Softmax(dim=1).to(device=device)



import src.measure
importlib.reload(src.measure)
from src.measure import compute_measures,eval_soundness_completeness, eval_soundness_completeness_class_specific, plot_parameters, plot_parameters_detailed, precision_bound



def file_name_fn(model_name, model_type, gamma, mask_size, use_NN_merlin, use_NN_morgana, seed):
    if model_name == 'arthur':
        return main_direc + f"trained_models_merlin_only/merlin_only_{model_type}_{mask_size}_trained_{int(use_NN_merlin)}_s{seed}.pt"
    elif model_name == 'merlin':
        return main_direc + f"trained_models_merlin_only/merlin_only_merlin_{model_type}_{mask_size}_trained_{int(use_NN_merlin)}_s{seed}.pt"
    else:
        raise ValueError("incorrect inputs for file name function")


P_e = {}
fano_vio = {}
avg_prec = {}
H_cond = {}
n_ipf = {}

for seed in seeds:
    P_e[seed], fano_vio[seed], avg_prec[seed], H_cond[seed], n_ipf[seed] = compute_measures(file_name_fn, trainLoader, testLoader, gamma, 
                                                                              masksizeList, use_NN_merlin, use_NN_morgana, params, 
                                                                              model_type = model_type, seed=seed)

P_e2 = {}
fano_vio2 = {}
avg_prec2 = {}
H_cond2 = {}
n_ipf2 = {} 

for ms in masksizeList:
    P_e2[ms] = []
    fano_vio2[ms] = []
    avg_prec2[ms] = []
    H_cond2[ms] = []
    n_ipf2[ms] = []
    
    for seed in seeds:
        P_e2[ms].append(P_e[seed][ms]) 
        fano_vio2[ms].append(fano_vio[seed][ms]) 
        avg_prec2[ms].append(avg_prec[seed][ms]) 
        H_cond2[ms].append(H_cond[seed][ms]) 
        n_ipf2[ms].append(n_ipf[seed][ms]) 

P_e_m = []
fano_vio_m = []
avg_prec_m = []
H_cond_m = []
n_ipf_m = []

P_e_s = []
fano_vio_s = []
avg_prec_s = []
H_cond_s = []
n_ipf_s = []

for ms in masksizeList:
    P_e_m.append(np.mean(P_e2[ms]))
    fano_vio_m.append(np.mean(fano_vio2[ms]))
    avg_prec_m.append(np.mean(avg_prec2[ms]))
    H_cond_m.append(np.mean(H_cond2[ms]))
    n_ipf_m.append(np.mean(n_ipf2[ms]))
    
    P_e_s.append(np.std(P_e2[ms]))
    fano_vio_s.append(np.std(fano_vio2[ms]))
    avg_prec_s.append(np.std(avg_prec2[ms]))
    H_cond_s.append(np.std(H_cond2[ms]))
    n_ipf_s.append(np.std(n_ipf2[ms]))


fig, ax = plt.subplots()

xvals = [log2(i) for i in masksizeList]

ax.errorbar(xvals, P_e_m, yerr=P_e_s, label='$P_{\mathrm{e}}$', linewidth=4, linestyle='-.', color='r', capsize=10.0, capthick=3)

ax.fill_between(xvals, np.array(P_e_m)-np.array(P_e_s), np.array(P_e_m)+np.array(P_e_s), alpha=0.3)

ax.errorbar(xvals, [-min(i,0) for i in fano_vio_m], label='$I_{coop}$', linewidth=4, color='g')
ax.errorbar(xvals, H_cond_m, yerr=H_cond_s, label='$H(C(Y)|Y \in M(X))$', linewidth=4, linestyle='--',color='b', capsize=10.0, capthick=3)

ax.fill_between(xvals, np.array(H_cond_m)-np.array(H_cond_s), np.array(H_cond_m)+np.array(H_cond_s), alpha=0.3)

plt.legend()

plt.grid()

plt.savefig(main_direc + f"./results_merlin_only/classes{classes_str}_adversarial_{model_names}_ticks.pdf", transparent=True, bbox_inches='tight',pad_inches = 0.015)


