import torch
import torch.nn as nn
from torchvision import models

def get_model(model_type='mlp', nclasses=10):
    if model_type == 'mlp':
        model = torch.nn.Sequential(
            torch.nn.Flatten(),
            torch.nn.Linear(28*28, 1024),
            torch.nn.ReLU(),
            torch.nn.Linear(1024, 1024),
            torch.nn.ReLU(),
            torch.nn.Linear(1024, 1 if nclasses==2 else nclasses),
        )
    elif model_type == 'conv':
        class SimpleCNN(torch.nn.Module):
            def __init__(self):
                super(SimpleCNN, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=2),
                    torch.nn.Dropout(p=0.25)
                )

                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(9216, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                x = torch.unsqueeze(x, 1)
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        
        model = SimpleCNN()
    elif model_type == 'conv_sq':
        # Input size: 56x56
        class CNN_SQ(torch.nn.Module):
            def __init__(self):
                super(CNN_SQ, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=5),
                    torch.nn.Dropout(p=0.25)
                )

                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(6400, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                x = torch.unsqueeze(x, 1)
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        
        model = CNN_SQ()
    elif model_type == 'conv_sq2':
        # Input size: 56x56
        class CNN_SQ2(torch.nn.Module):
            def __init__(self):
                super(CNN_SQ2, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, stride=2, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=2),
                    torch.nn.Dropout(p=0.25)
                )

                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(4992, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                x = torch.unsqueeze(x, 1)
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        model = CNN_SQ2()
    elif model_type == 'conv_sq3':
        class CNN_SQ(torch.nn.Module):
            def __init__(self):
                super(CNN_SQ, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=2),
                    torch.nn.Dropout(p=0.25)
                )

                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(9216, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                x = torch.unsqueeze(x, 1)
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        
        model = CNN_SQ()
    elif model_type == 'convSTL10':
        class SimpleCNN(torch.nn.Module):
            def __init__(self):
                super(SimpleCNN, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=3, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=2),
                    torch.nn.Dropout(p=0.25)
                )

                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(135424, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        
        model = SimpleCNN()
    elif model_type == 'conv_CIFAR10':
        class SimpleCNN(torch.nn.Module):
            def __init__(self):
                super(SimpleCNN, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=3, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=2),
                    torch.nn.Dropout(p=0.25)
                )
                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(10816, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                #x = torch.unsqueeze(x, 1) #This is only needed if there is no channel layer already
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        
        model = SimpleCNN()
    elif model_type == 'conv_spinal':
        model_ft = models.wide_resnet101_2(pretrained=True)
        num_ftrs = model_ft.fc.in_features

        half_in_size = round(num_ftrs/2)
        layer_width = 512
        Num_class=10

        class SpinalNet(torch.nn.Module):
            def __init__(self):
                super(SpinalNet, self).__init__()

                self.fc_spinal_layer1 = nn.Sequential(
                    nn.Dropout(p = 0.5), nn.Linear(half_in_size, layer_width),
                    nn.BatchNorm1d(layer_width), nn.ReLU(inplace=True),)
                self.fc_spinal_layer2 = nn.Sequential(
                    nn.Dropout(p = 0.5), nn.Linear(half_in_size+layer_width, layer_width),
                    nn.BatchNorm1d(layer_width), nn.ReLU(inplace=True),)
                self.fc_spinal_layer3 = nn.Sequential(
                    nn.Dropout(p = 0.5), nn.Linear(half_in_size+layer_width, layer_width),
                    nn.BatchNorm1d(layer_width), nn.ReLU(inplace=True),)
                self.fc_spinal_layer4 = nn.Sequential(
                    nn.Dropout(p = 0.5), nn.Linear(half_in_size+layer_width, layer_width),
                    nn.BatchNorm1d(layer_width), nn.ReLU(inplace=True),)
                self.fc_out = nn.Sequential(
                    nn.Dropout(p = 0.5), nn.Linear(layer_width*4, Num_class),)



            def forward(self, x):
                x1 = self.fc_spinal_layer1(x[:, 0:half_in_size])
                x2 = self.fc_spinal_layer2(torch.cat([ x[:,half_in_size:2*half_in_size], x1], dim=1))
                x3 = self.fc_spinal_layer3(torch.cat([ x[:,0:half_in_size], x2], dim=1))
                x4 = self.fc_spinal_layer4(torch.cat([ x[:,half_in_size:2*half_in_size], x3], dim=1))


                x = torch.cat([x1, x2], dim=1)
                x = torch.cat([x, x3], dim=1)
                x = torch.cat([x, x4], dim=1)


                x = self.fc_out(x)
                return x
        model = SpinalNet()
          
    elif model_type == 'conv_deeper':
        class DeeperCNN(torch.nn.Module):
            def __init__(self):
                super(DeeperCNN, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=2),
                    #torch.nn.Dropout(p=0.25)
                )

                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(7744, 1024),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Linear(1024, 128),
                    torch.nn.ReLU(inplace=True),
                    #torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )

            def forward(self, x):
                """Perform forward."""
                x = torch.unsqueeze(x, 1)
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        
        model = DeeperCNN()
    
    elif model_type == 'conv_sq':
        class CNN_SQ(torch.nn.Module):
            def __init__(self):
                super(CNN_SQ, self).__init__()
                self.conv_layer = torch.nn.Sequential(
                    # Conv Layer block 1
                    torch.nn.Conv2d(in_channels=1, out_channels=32, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=0),
                    torch.nn.ReLU(inplace=True),
                    #torch.nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, padding=0),
                    #torch.nn.ReLU(inplace=True),
                    torch.nn.MaxPool2d(kernel_size=5),
                    torch.nn.Dropout(p=0.25)
                )
                self.fc_layer = torch.nn.Sequential(
                    torch.nn.Linear(6400, 128),
                    torch.nn.ReLU(inplace=True),
                    torch.nn.Dropout(p=0.5),
                    torch.nn.Linear(128, 1 if nclasses==2 else nclasses)
                )
            def forward(self, x):
                """Perform forward."""
                x = torch.unsqueeze(x, 1)
                x = self.conv_layer(x)
                x = x.view(x.size(0), -1)
                x = self.fc_layer(x)
                return x
        model = CNN_SQ()
        
    return model