from src.utilities import Averager
from src.arthur_merlin_networks import opt_mask_NN_FW
from src.arthur_merlin_optimizers import opt_mask
import torch
import numpy as np

def H_b(p):
    if p==0 or p==1:
        return 0
    else:
        return -p*np.log(p) - (1-p)*np.log(1-p)

def fano_violation(P_error, H_cond, nclasses):
    return H_b(P_error) + P_error*(np.log(nclasses-1)) - H_cond

def calc_entropy(p):
    return -torch.xlogy(p,p).sum(dim=1)

def calc_precision(probs, y_target):
    return torch.gather(probs,1,y_target.unsqueeze(1))

def get_completeness(model, loader, use_NN, merlin_model = None, mask_size = None, max_iter=1500, params = None):
    assert mask_size
    
    device = params['device']
    completeness = Averager()
    #count = 0
    for x_input, y_target in loader:
        x_image = x_input.select(dim=1, index=0)
        x_image, y_target = x_image.to(device), y_target.to(device)
        
        mask = x_input.select(dim=1, index=1)
        if use_NN == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=mask_size, mask=mask, 
                                                                stoptol=1e-5, max_iter=5, mask_model = merlin_model,
                                                                mask_model_opt = None, params=params, test_mode = True)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=mask_size, mask=mask,
                                                          max_iter=max_iter, params = params)
        completeness.add(merlin_metrics['accuracy'])
    return completeness.result()

def get_completeness_STL10(model, loader, use_NN, merlin_model = None, mask_size = None, max_iter=1500, params = None, channels=3):
    assert mask_size
    
    device = params['device']
    completeness = Averager()
    #count = 0
    for x_input, y_target in loader:
        x_image = x_input[:,0:channels,:,:]
        x_image, y_target = x_image.to(device), y_target.to(device)
        
        mask = x_input.select(dim=1, index=1)
        if use_NN == True:
            x_merlin, s_merlin, merlin_metrics = opt_mask_NN_FW_STL10(model, x_image, y_target, 'merlin', k=mask_size, mask=mask, 
                                                                stoptol=1e-5, max_iter=5, mask_model = merlin_model,
                                                                mask_model_opt = None, params=params, test_mode = True)
        else:
            x_merlin, s_merlin, merlin_metrics = opt_mask(model, x_image, y_target, 'merlin', k=mask_size, mask=mask,
                                                          max_iter=max_iter, params = params)
        completeness.add(merlin_metrics['accuracy'])
    return completeness.result()

def get_class_count(x_image, mask, fixedLoader, diffTol=1E-7, params=None):
    
    device = params['device']
    # Mask and make space for comparison batch dimension
    mask = mask.unsqueeze(1)
    x_image = x_image.unsqueeze(1)
    x_masked = (mask*x_image)
    nclasses = params['nclasses']
    class_count = torch.zeros(mask.size(0), nclasses).to(device)
    
    for x_comp,y_comp in fixedLoader:
        x_comp_image = x_comp.select(dim=1, index=0)
        x_comp_image, y_comp = x_comp_image.to(device), y_comp.to(device)
        
        #Mask and make space for feature batch dimension
        x_comp_image = x_comp_image.unsqueeze(dim = 0)
        x_comp_masked = mask*x_comp_image
        
        # print(x_masked.shape, x_comp_masked.shape)
        
        #Calculate difference

        feature_deviation = torch.norm(x_masked-x_comp_masked, p=2, dim=(2,3))
        feature_boolmask = feature_deviation < diffTol
        
        for label in range(nclasses):
            
            # create label mask to compare to feature 
            label_boolmask = (y_comp==label).unsqueeze(dim=0)
            
            feature_label_boolmask = torch.logical_and(feature_boolmask, label_boolmask)
            del label_boolmask
            
            #counts incidence of label for every feature
            incidence_count = feature_label_boolmask.sum(dim=1)
            del feature_label_boolmask
            
            class_count[:,label] += incidence_count

    return class_count

def get_class_count_STL10(x_image, mask, fixedLoader, diffTol=1E-7, params=None, channels=3):
    

    device = params['device']
    # Mask and make space for comparison batch dimension
    mask = mask.unsqueeze(1)
    x_image = x_image.unsqueeze(1)
    x_masked = (mask*x_image)
    nclasses = params['nclasses']
    class_count = torch.zeros(mask.size(0), nclasses).to(device)
    
    for x_comp,y_comp in fixedLoader:
        x_comp_image = x_comp[:,0:channels,:,:]
        x_comp_image, y_comp = x_comp_image.to(device), y_comp.to(device)
        
        #Mask and make space for feature batch dimension
        x_comp_image = x_comp_image.unsqueeze(dim = 0)
        x_comp_masked = mask*x_comp_image
        
        # print(x_masked.shape, x_comp_masked.shape)
        
        #Calculate difference

        feature_deviation = torch.norm(x_masked-x_comp_masked, p=2, dim=(2,3,4))
        feature_boolmask = feature_deviation < diffTol
        
        print(feature_deviation.shape)
        
        for label in range(nclasses):
            
            # create label mask to compare to feature 
            label_boolmask = (y_comp==label).unsqueeze(dim=0)
            print(label_boolmask.shape)
            
            feature_label_boolmask = torch.logical_and(feature_boolmask, label_boolmask)
            del label_boolmask
            
            #counts incidence of label for every feature
            incidence_count = feature_label_boolmask.sum(dim=1)
            del feature_label_boolmask
            
            class_count[:,label] += incidence_count

    return class_count

def get_H_cond(model, testLoader, fixedLoader, mask_size, diffTol, params, merlin_model=None,use_NN = False):
    
    device = params['device']
    mean_H_cond = Averager()
    mean_images_per_feature = Averager()
    avg_precision = Averager()

    for (x_input, y_target) in testLoader:
        
        x_image = x_input.select(dim=1, index=0)
        mask = x_input.select(dim=1, index=1)
        x_image, y_target, mask = x_image.to(device), y_target.to(device), mask.to(device)
        
        if use_NN == False:
            feature, mask, metrics = opt_mask(model, x_image, y_target, 'merlin', k=mask_size, mask=mask, params=params)
        else:
            feature, mask, metrics = opt_mask_NN_FW(model, x_image, y_target, 'merlin', k=mask_size, mask=mask, 
                                                                stoptol=1e-5, max_iter=5, mask_model = merlin_model,
                                                                mask_model_opt = None, params=params, test_mode = True)
        
        class_count = get_class_count(x_image, mask, fixedLoader, diffTol, params)
        incidence_count = class_count.sum(dim = 1, keepdim=True)
        class_prob = class_count/incidence_count
        
        # Average Precision
        avg_precision.add(calc_precision(class_prob, y_target).mean())
        
        # Calculate conditional entropy
        H_cond = calc_entropy(class_prob)
        mean_H_cond.add(H_cond.mean())
        
        # Save mean number of found images for a feature
        mean_images_per_feature.add(incidence_count.mean())
        
    return avg_precision.result().item(), mean_H_cond.result().item(), mean_images_per_feature.result().item()


def get_H_cond_STL10(model, testLoader, fixedLoader, mask_size, diffTol, params, merlin_model=None,use_NN = False, channels=3):
    
    device = params['device']
    mean_H_cond = Averager()
    mean_images_per_feature = Averager()
    avg_precision = Averager()

    for (x_input, y_target) in testLoader:
        
        x_image = x_input[:,0:channels,:,:]
        mask = x_input.select(dim=1, index=1)
        x_image, y_target, mask = x_image.to(device), y_target.to(device), mask.to(device)
        
        if use_NN == False:
            feature, mask, metrics = opt_mask(model, x_image, y_target, 'merlin', k=mask_size, mask=mask, params=params)
        else:
            feature, mask, metrics = opt_mask_NN_FW_STL10(model, x_image, y_target, 'merlin', k=mask_size, mask=mask, 
                                                                stoptol=1e-5, max_iter=5, mask_model = merlin_model,
                                                                mask_model_opt = None, params=params, test_mode = True)
        
        class_count = get_class_count_STL10(x_image, mask, fixedLoader, diffTol, params)
        incidence_count = class_count.sum(dim = 1, keepdim=True)
        class_prob = class_count/incidence_count
        
        # Average Precision
        avg_precision.add(calc_precision(class_prob, y_target).mean())
        
        # Calculate conditional entropy
        H_cond = calc_entropy(class_prob)
        mean_H_cond.add(H_cond.mean())
        
        # Save mean number of found images for a feature
        mean_images_per_feature.add(incidence_count.mean())
        
    return avg_precision.result().item(), mean_H_cond.result().item(), mean_images_per_feature.result().item()