from multiprocessing.sharedctypes import Value

import torch


def custom_criterion(
    logits: torch.Tensor, y_target: torch.Tensor, reduction: str = "mean"
) -> torch.Tensor:
    correct_logits = torch.gather(logits, 1, y_target.unsqueeze(1))
    idk_logits = logits[:, 2].unsqueeze(1)

    diff = -torch.abs(correct_logits - idk_logits)

    if reduction == "mean":
        correction_term = -torch.log(1 + torch.exp(diff)).mean()
    elif reduction == "sum":
        correction_term = -torch.log(1 + torch.exp(diff)).sum()
    elif reduction == "none":
        correction_term = -torch.log(1 + torch.exp(diff)).squeeze()
    else:
        raise ValueError(f"unexpected value for reduction, got {reduction}")

    new_logits = torch.cat((correct_logits, idk_logits), 1)
    y_new = torch.clone(y_target)
    y_new[torch.argmax(new_logits, dim=1) == 1] = 2
    new_crit = torch.nn.CrossEntropyLoss(reduction=reduction)
    loss = -(new_crit(logits, y_new) + correction_term)

    return loss


def add_probs_criterion(logits: torch.Tensor, y_target: torch.Tensor) -> torch.Tensor:
    probs = torch.softmax(logits, dim=1)
    correct_probs = torch.gather(probs, 1, y_target.unsqueeze(1))
    idk_probs = probs[:, 2].unsqueeze(1)
    loss = (correct_probs + idk_probs).mean()

    return loss
